# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Heightmap Export
 -------------------
 Release Date: 2019-01-24
 Nathaniel Klumb
 Email: heightmap@clayjar.com
 git SHA: $Format:%H$
 ***************************************************************************/

 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from math import sin,cos,atan2,sqrt,radians

from PyQt4 import QtCore, QtGui
from PyQt4.QtGui import QDialog, QMessageBox, QColor, QFileDialog, QApplication, QCursor, QDoubleValidator
from PyQt4.QtCore import Qt, QLocale
from qgis.gui import QgsRubberBand, QgsMapTool, QgsMessageBar
from osgeo import gdal
#import struct

from .HeightmapExportDialog_base import Ui_HeightmapExportDialog, _fromUtf8
from qgis._core import QgsPoint, QgsRectangle, QgsMapLayerRegistry, QgsGeometry, QgsCoordinateTransform, QgsCoordinateReferenceSystem

class HeightmapExportDialog(QtGui.QDialog, Ui_HeightmapExportDialog):
    SIZE_DECIMAL_DIGITS = 3
    SCALE_DECIMAL_DIGITS = 5
    MAXIMUM_ALLOWED_VALUE = 6371000000
    
    layers = []
    layer_index = -1
    layer = None
    extent = None
    #latn = 0
    #lats = 0
    #lonw = 0
    #lone = 0
    height = 0
    width = 0
    zmin = 0
    zmax = 0
    src = None

    def __init__(self, iface):
        """Constructor."""
        QDialog.__init__(self)
        self.ui = Ui_HeightmapExportDialog()
        self.ui.setupUi(self)
        self.iface = iface
        self.canvas = iface.mapCanvas()

        # Get the list of valid raster layers and set up the combobox.
        # The first valid raster layer will be selected automatically.
        self.list_layers()
        #self.ui.LayerComboBox.activated[unicode].connect(self.get_layer)
        self.ui.LayerComboBox.activated.connect(self.set_active_layer)
        # endregion

        # region Dimension actions
        validDecimals = QDoubleValidator(0.1 ** self.SIZE_DECIMAL_DIGITS, self.MAXIMUM_ALLOWED_VALUE, self.SIZE_DECIMAL_DIGITS)
        self.ui.ModelHeight.setValidator(validDecimals)
        self.ui.ModelHeight.textEdited.connect(self.update_from_model_height)
        self.ui.ModelWidth.setValidator(validDecimals)
        self.ui.ModelWidth.textEdited.connect(self.update_from_model_width)
        self.ui.ModelDepth.setValidator(validDecimals)
        self.ui.ModelDepth.textEdited.connect(self.update_from_model_depth)
        self.ui.ScaleDepth.setValidator(validDecimals)
        self.ui.ScaleDepth.textEdited.connect(self.update_from_scale_depth)

        validScales = QDoubleValidator(0.1 ** self.SCALE_DECIMAL_DIGITS, self.MAXIMUM_ALLOWED_VALUE, self.SCALE_DECIMAL_DIGITS)
        self.ui.ScaleFactor.setValidator(validScales)
        self.ui.ScaleFactor.textEdited.connect(self.update_from_scale_factor)
        self.ui.Resolution.setValidator(validScales)
        self.ui.Resolution.textEdited.connect(self.update_from_resolution)
        
        validIntegers = QDoubleValidator(1, self.MAXIMUM_ALLOWED_VALUE, 0)
        self.ui.ImageHeight.setValidator(validScales)
        self.ui.ImageHeight.textEdited.connect(self.update_from_image_height)
        self.ui.ImageWidth.setValidator(validScales)
        self.ui.ImageWidth.textEdited.connect(self.update_from_image_width)
        # endregion

        # region Cancel, export, print buttons
        self.ui.CancelDialog.clicked.connect(self.reject)
        self.ui.ExportHeightmap.clicked.connect(self.export_heightmap)
        close = QtGui.QAction(self)
        self.connect(close, QtCore.SIGNAL('clicked()'), self.reject)
        # endregion

    def check_ready(self):
      self.ui.CancelDialog.setText('Cancel')
      try:
          if float(self.ui.ModelHeight.text()) and float(self.ui.ModelWidth.text()) and int(self.ui.ImageHeight.text()) and int(self.ui.ImageWidth.text()):
              self.ui.ExportHeightmap.setEnabled(True)
          else:
              self.ui.ExportHeightmap.setEnabled(False)
      except:
          self.ui.ExportHeightmap.setEnabled(False)
        
    def scale_factor(self):
      try:
        sfactor = float(self.ui.ScaleFactor.text())
      except:
        sfactor = 1.0
        self.ui.ScaleFactor.setText('1')
      return sfactor
    
    def update_scales(self, also_resolution = True):
        try:
          mheight = float(self.ui.ModelHeight.text())
          sdepth = float(self.ui.ScaleDepth.text())
        except:
          mheight = False
          sdepth = False
        if mheight and sdepth:
            self.ui.HorizontalScale.setText('%s m/mm (%s:1)' % (str(round(self.height/mheight,self.SCALE_DECIMAL_DIGITS)),QLocale().toString(float(self.height/mheight*1000),'f',0)))
            self.ui.VerticalScale.setText('%s m/mm (%s:1)' % (str(round((self.zmax - self.zmin)/sdepth,self.SCALE_DECIMAL_DIGITS)),QLocale().toString(float((self.zmax-self.zmin)/sdepth*1000),'f',0)))
            if also_resolution:
                try:
                  iheight = float(self.ui.ImageHeight.text())
                except:
                  iheight = False
                if iheight:
                    self.ui.Resolution.setText(str(round(mheight/iheight,self.SCALE_DECIMAL_DIGITS)))
        self.check_ready()
    
    def update_from_model_height(self, mheight):
        try:
          mheight = float(mheight)
        except:
          return
        if not self.height:
          return
        self.ui.ModelWidth.setText(str(round(mheight/self.height*self.width,self.SIZE_DECIMAL_DIGITS)))
        self.ui.ModelDepth.setText(str(round(mheight/self.height*(self.zmax - self.zmin),self.SIZE_DECIMAL_DIGITS)))
        self.ui.ScaleDepth.setText(str(round(mheight/self.height*(self.zmax - self.zmin)*self.scale_factor(),self.SIZE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_model_width(self, mwidth):
        try:
          mwidth = float(mwidth)
        except:
          return
        if not self.width:
          return
        self.ui.ModelHeight.setText(str(round(mwidth/self.width*self.height,self.SIZE_DECIMAL_DIGITS)))
        self.ui.ModelDepth.setText(str(round(mwidth/self.width*(self.zmax - self.zmin),self.SIZE_DECIMAL_DIGITS)))
        self.ui.ScaleDepth.setText(str(round(mwidth/self.width*(self.zmax - self.zmin)*self.scale_factor(),self.SIZE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_model_depth(self, mdepth):
        try:
          mdepth = float(mdepth)
        except:
          return
        if (self.zmax - self.zmin) == 0:
          return
        self.ui.ModelHeight.setText(str(round(mdepth/(self.zmax - self.zmin)*self.height,self.SIZE_DECIMAL_DIGITS)))
        self.ui.ModelWidth.setText(str(round(mdepth/(self.zmax - self.zmin)*self.width,self.SIZE_DECIMAL_DIGITS)))
        try:
          sfactor = float(self.ui.ScaleFactor.text())
        except:
          sfactor = 1.0
          self.ui.ScaleFactor.setText('1')
        self.ui.ScaleDepth.setText(str(round(mdepth*sfactor,self.SIZE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_scale_depth(self, sdepth):
        try:
          mdepth = float(self.ui.ModelDepth.text())
          sdepth = float(sdepth)
        except:
          return
        if not mdepth or not sdepth:
          return
        self.ui.ScaleFactor.setText(str(round(sdepth/mdepth,self.SCALE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_scale_factor(self, sfactor):
        try:
          mdepth = float(self.ui.ModelDepth.text())
          sfactor = float(sfactor)
        except:
          return
        if not mdepth or not sfactor:
          return
        self.ui.ScaleDepth.setText(str(round(mdepth*sfactor,self.SCALE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_resolution(self, res):
        try:
            mheight = float(self.ui.ModelHeight.text())
            res = float(res)
        except:
          return
        if not mheight or not res:
          return
        self.ui.ImageHeight.setText('%i' % round(mheight/res))
        self.ui.ImageWidth.setText('%i' % round(mheight/self.height*self.width/res))
        self.update_scales(False)

    def update_from_image_height(self, iheight):
        try:
          mheight = float(self.ui.ModelHeight.text())
          iheight = round(float(self.ui.ImageHeight.text()))
          #self.ui.ImageHeight.setText(str(iheight))
        except:
          return
        if not mheight or not iheight:
          return
        self.ui.ImageWidth.setText('%i' % round(iheight/self.height*self.width))
        #self.ui.Resolution.setText(str(round(mheight/iheight,self.SCALE_DECIMAL_DIGITS)))
        self.update_scales()

    def update_from_image_width(self, iwidth):
        try:
          mwidth = float(self.ui.ModelWidth.text())
          iwidth = float(self.ui.ImageWidth.text())
          #self.ui.ImageWidth.setText(str(iwidth))
        except:
          return
        if not mwidth or not iwidth:
          return
        self.ui.ImageHeight.setText('%i' % round(iwidth/self.width*self.height))
        #self.ui.Resolution.setText(str(round(mwidth/iwidth,self.SCALE_DECIMAL_DIGITS)))
        self.update_scales()
    
    def set_active_layer(self, layer_index):
        if self.layer_index != layer_index:
            self.layer_index = layer_index
            self.reset_dialog()
            self.layer = QgsMapLayerRegistry.instance().mapLayer(self.layers[layer_index]['LayerID'])
            self.get_dimensions()
            self.draw_box()

    def list_layers(self):
        self.layers = []
        self.ui.LayerComboBox.clear()
        maplayers = QgsMapLayerRegistry.instance().mapLayers()
        for layerID,layer in maplayers.iteritems():
            if (layer.type() == 1) and layer.isValid(): #(LayerType == RasterLayer) and valid
                self.layers += [{ 'LayerID': layerID, 'LayerName': layer.name() }]
                self.ui.LayerComboBox.addItem(layer.name())
        if len(self.layers):
            self.set_active_layer(0)
        else:
            QMessageBox.warning(self, "Nothing to do here.", "No valid raster layers loaded.")
            self.reject()

    
    def reset_dialog(self):
        self.ui.ModelHeight.clear()
        self.ui.ModelWidth.clear()
        self.ui.ModelDepth.clear()
        self.ui.ScaleDepth.clear()
        self.ui.ScaleFactor.setText('1')
        self.ui.Resolution.clear()
        self.ui.ImageHeight.clear()
        self.ui.ImageWidth.clear()
        self.ui.HorizontalScale.setText('')
        self.ui.VerticalScale.setText('')
        if self.extent:
            self.canvas.scene().removeItem(self.extent)
            self.extent = None

    def get_dimensions(self):
        if (self.layer is None):
          return
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        #layer = get_layer(layers[0].name())
        extent = self.layer.extent()
        source = self.layer.crs()
        wgs84 = QgsCoordinateReferenceSystem("EPSG:4326")
        transform = QgsCoordinateTransform(source,wgs84)
      
        rect = transform.transform(extent)
        latn = rect.yMaximum()
        lats = rect.yMinimum()
        lonw = rect.xMinimum()
        lone = rect.xMaximum()
        height = 6371000 * 2 * atan2(sqrt((sin((radians(lats)-radians(latn))/2))**2),sqrt(1-(sin((radians(lats)-radians(latn))/2))**2))
        width = 6371000 * 2 * atan2(sqrt((cos((radians(lats)+radians(latn))/2))**2 * (sin((radians(lone)-radians(lonw))/2))**2),
                                    sqrt(1-((cos((radians(lats)+radians(latn))/2))**2 * (sin((radians(lone)-radians(lonw))/2))**2)))

        #provider = layer.dataProvider()
        #path = provider.dataSourceUri()
        #path_layer = path.split('|')
        #src = gdal.Open(path_layer[0])
        self.src = gdal.Open(self.layer.dataProvider().dataSourceUri().split('|')[0])
        (zmin,zmax) = self.src.GetRasterBand(1).ComputeRasterMinMax(False)
    
        #self.latn = latn
        #self.lats = lats
        #self.lonw = lonw
        #self.lone = lone
        self.height = height
        self.width = width
        self.zmin = zmin
        self.zmax = zmax
        
        self.ui.MapHeight.setText(QLocale().toString(float(height),'f',self.SIZE_DECIMAL_DIGITS))
        self.ui.MapWidth.setText(QLocale().toString(float(width),'f',self.SIZE_DECIMAL_DIGITS))
        self.ui.MapHighest.setText(QLocale().toString(float(zmax),'f',self.SIZE_DECIMAL_DIGITS))
        self.ui.MapLowest.setText(QLocale().toString(float(zmin),'f',self.SIZE_DECIMAL_DIGITS))
        QApplication.restoreOverrideCursor()
        
    # region Extension functions
    def draw_box(self):
        rec = self.layer.extent()
        x_max = rec.xMaximum()
        y_min = rec.yMinimum()
        x_min = rec.xMinimum()
        y_max = rec.yMaximum()

        if self.extent:
            self.canvas.scene().removeItem(self.extent)
            self.extent = None
        self.extent = QgsRubberBand(self.canvas, True)
        points = [QgsPoint(x_max,y_min), QgsPoint(x_max,y_max), QgsPoint(x_min,y_max), QgsPoint(x_min,y_min), QgsPoint(x_max,y_min)]
        self.extent.setToGeometry(QgsGeometry.fromPolyline(points), None)
        self.extent.setColor(QColor(255, 25, 25, 255))
        self.extent.setWidth(4)
        self.extent.setLineStyle(Qt.PenStyle(Qt.DashLine))
        self.canvas.refresh()
    # endregion

    def export_heightmap(self):
        try:
            suggestion = self.layer.dataProvider().dataSourceUri().split('|')[0].rsplit('.',1)[0] + '_heightmap.png'
        except:
            suggestion = self.layers[self.layer_index]['LayerName'] + '_heightmap.png'
        filenames = QFileDialog.getSaveFileNameAndFilter(self, 'Export to PNG', suggestion, filter="16-bit PNG Heightmap (*.png)")
        png_file = filenames[0]
        if png_file:
            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
            ds = gdal.Translate(png_file, 
                                self.src,
                                format = 'PNG',
                                outputType = gdal.GDT_UInt16,
                                width = int(self.ui.ImageWidth.text()),
                                height = int(self.ui.ImageHeight.text()),
                                scaleParams = [[self.zmin,self.zmax,0,65535]]
                               )
            self.ui.CancelDialog.setText('Close')
            QApplication.restoreOverrideCursor()
            QMessageBox.information(self, "Heightmap Export", "Save complete.")
